# stepic - Python image steganography
# Copyright (C) 2007 Lenny Domnitser
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

'''Python image steganography

Stepic hides arbitrary data inside PIL images.

Besides Python itself, Stepic only depends on the Python Image Library
(apt: python-imaging, web: <http://www.pythonware.com/products/pil/>).

Usually, the best way to use stepic is to call encode(image, data) and
decode(image). If one source image is used to produce several data-enclosing
images, then you may wish to access the backend to encode and decode, which
is Steganographer.
'''

__author__ = 'Lenny Domnitser <http://domnit.org/>'
__version__ = '0.2.1'


import Image


__all__ = ('Steganographer', 'encode', 'decode')


class Steganographer:
    '''encodes and decodes data in an image

    This may be useful if one source image is to be used to produce several
    data-enclosing images.
    '''

    def __init__(self, image):
        if image.mode not in ('RGB', 'RGBA'):
            raise TypeError('image is not RGB or RGBA')

        self.image = image
        self.width = image.size[0]
        self.height = image.size[1]


    def _get_next_pixel_location(self, pixel_location):
        '''given (x, y), finds next (x, y), searching from left to right,
        advancing to the next row as needed
        '''

        x, y = pixel_location
        if x + 1 == self.width:
            return (0, y + 1)
        return (x + 1, y)


    def encode(self, data):
        '''generates a new image with hidden data
        '''
        if len(data) == 0:
            raise ValueError('data is empty')
        if len(data) > self.width * self.height // 3:
            raise ValueError('data is too large for image')
    
        image = self.image.copy()

        pixel_location = (0, 0)
        data_index = 0
        stop_index = len(data) - 1

        while data_index <= stop_index:
            pixel1_at = pixel_location
            pixel2_at = self._get_next_pixel_location(pixel1_at)
            pixel3_at = self._get_next_pixel_location(pixel2_at)
            pixel_location = self._get_next_pixel_location(pixel3_at)

            flat_pixel_data = self.image.getpixel(pixel1_at)[:3] + \
                              self.image.getpixel(pixel2_at)[:3] + \
                              self.image.getpixel(pixel3_at)[:3]
            flat_pixel_data = [value & -2 for value in flat_pixel_data]

            byte = ord(data[data_index])
            for c in xrange(-2, -10, -1):
                flat_pixel_data[c] |= byte & 1
                byte >>= 1

            if data_index == stop_index:
                flat_pixel_data[-1] |= 1

            image.putpixel(pixel1_at, tuple(flat_pixel_data[0:3]))
            image.putpixel(pixel2_at, tuple(flat_pixel_data[3:6]))
            image.putpixel(pixel3_at, tuple(flat_pixel_data[6:9]))

            data_index += 1

        return image


    def decode(self):
        '''finds data in self's image
        '''
        data = []

        pixel_location = (0, 0)

        while True:
            pixel1_at = pixel_location
            pixel2_at = self._get_next_pixel_location(pixel1_at)
            pixel3_at = self._get_next_pixel_location(pixel2_at)
            pixel_location = self._get_next_pixel_location(pixel3_at)

            flat_pixel_data = self.image.getpixel(pixel1_at)[:3] + \
                              self.image.getpixel(pixel2_at)[:3] + \
                              self.image.getpixel(pixel3_at)[:3]

            byte = 0
            for c in xrange(7):
                byte |= flat_pixel_data[c] & 1
                byte <<= 1
            byte |= flat_pixel_data[7] & 1

            byte = chr(byte)
            data.append(byte)

            if flat_pixel_data[-1] & 1:
                break

        data = ''.join(data)
        return data


def encode(image, data):
    '''generates an image with hidden data, starting with an existing image
    and arbitrary data
    '''
    return Steganographer(image).encode(data)


def decode(image):
    '''extracts data from an image
    '''
    return Steganographer(image).decode()
